using System;
using System.Text;
using System.Collections;
using System.Text.RegularExpressions;

namespace Danga 
{
    class HTTPHeaders 
    {
	private bool         complete = false;

	// 8bit preserving, HTTP standard
	private static Encoding encISO = Encoding.GetEncoding("ISO-8859-1");

	// common regexps
	private static Regex reHeader = new Regex(@"^(?<header>[\w\-]+):\s*(?<value>.*?)\r?\n$");
	private static Regex reFolded = new Regex(@"^(?<value>\s+.*?)\r?\n$");

	// last header we parsed
	private string lastHeader = null;

	private string requestLine = null;
	
	private Hashtable  headers;   // lowercase header -> comma-sep list of values
	private Hashtable  origcase;  // canonical header (lowercase) -> provided case
	private ArrayList  hdorder;     // order headers were received in, canonical values

	public HTTPHeaders () 
	{ 
	    headers  = new Hashtable();
	    origcase = new Hashtable();
	    hdorder  = new ArrayList();
	}

	public bool Complete {
	    get {
		return complete;
	    }
	}

	public string RequestLine {
	    get {
		return requestLine;
	    }
	    set {
		requestLine = value;
	    }
	}

	public bool ParseNewHeader (string s) 
	{
	    // it's an error to parse new headers even we've already read them all
	    if (complete)
		return false;

	    // make both LF and CRLF end the headers.
	    if (s == "\n" || s == "\r\n") {
		complete = true;
		return true;
	    }

	    Match m = reHeader.Match(s);
	    if (m.Success) {
		string hd = m.Groups["header"].ToString();
		string canHd = hd.ToLower();
		origcase[canHd] = hd;
		hd = canHd;

		if (headers[hd] == null) {
		    headers[hd] = m.Groups["value"].ToString();
		    hdorder.Add(hd);
		} else {
		    headers[hd] += ", " + m.Groups["value"].ToString();
		}
		lastHeader = hd;
		return true;
	    }

	    // must be a continuation from previous line (folded header)
	    if (lastHeader == null)
		return false;

	    m = reFolded.Match(s);
	    if (! m.Success)
		return false;

	    headers[lastHeader] += m.Groups["value"].ToString();
	    return true;
	}

	public void SetHeader (string header, string value) {
	    string hd = header.ToLower();
	    origcase[hd] = header;
	    if (headers[hd] == null)
		hdorder.Add(hd);
	    headers[hd] = value;
	}
	
	public string GetHeader (string header) 
	{
	    string canHd = header.ToLower();
	    return (string) headers[canHd];
	}

	public override string ToString () {
	    StringBuilder sb = new StringBuilder();
	    if (requestLine != null)
		sb.Append(requestLine);
	    foreach (string hd in hdorder) {
		sb.AppendFormat("{0}: {1}\r\n", origcase[hd], headers[hd]);
	    }
	    sb.Append("\r\n");
	    return sb.ToString();
	}

	// returns request as a byte array
	public byte[] GetBytes () 
	{
	    String s = ToString();
	    return encISO.GetBytes(s);
	}

    }
    
}
